/* This file is part of the Project Athena Zephyr Notification System.
 * It is one of the source files comprising zwgc, the Zephyr WindowGram
 * client.
 *
 *      Created by:     Marc Horowitz <marc@athena.mit.edu>
 *
 *      $Id: b8e78de10e3afabb13f07a6e559183540141a9e0 $
 *
 *      Copyright (c) 1989 by the Massachusetts Institute of Technology.
 *      For copying and distribution information, see the file
 *      "mit-copyright.h".
 */

#include <sysdep.h>

#if (!defined(lint) && !defined(SABER))
static const char rcsid_error_c[] = "$Id: b8e78de10e3afabb13f07a6e559183540141a9e0 $";
#endif

#include <zephyr/mit-copyright.h>

int error_code;
